#!/uns/bin/python

import os, time, select, string, tempfile, sys, traceback, types
import pygale, gale_pack, rsaref, sign, gale_env, engine

HOME = gale_env.get('FUGU_DIR')
DEBUG = 0
PRIVATE_MAGIC = 'h\023\000\001'
PRIVATE_MAGIC2 = 'h\023\000\003'

KEY_RESPONSE_CAT = '@%s/auth/key/%s/'
KEY_REQUEST_CAT = '@%s/auth/query/%s/'

NORMAL = 0
TRUSTED = 1

# Timeout for AKD requests (in milliseconds)
AKD_TIMEOUT = 30000
# Time between AKD requests for same key (in seconds)
AKD_INTERVAL = 60

class Collector:
	def __init__(self, inputs):
		self.__inputs = inputs
		self.__values = {}
		for item in inputs:
			self.__values[item] = None
	
	def setoutput(self, input, output):
		self.__values[input] = output
	
	def done(self):
		for item in self.__values.keys():
			if self.__values[item] is None:
				return 0
		return 1
	
	def getvalues(self):
		return map(lambda x, d=self.__values: d[x], self.__inputs)
	
class KeyCacheClass:
	def __init__(self):
		self.__keys = {}		# in-memory key cache
		self.__last_akd = {}	# Time we did last AKD request for key
		self.__keychain = {}	# List of callbacks waiting for a key

	def get_key(self, name):
		# TODO: check for newer on-disk version
		if self.__keys.has_key(name):
			return self.__keys[name]
		else:
			return None

	def import_privkey(self, key):
		# _ga_import_priv
		(magic, key) = gale_pack.pop_data(key, len(PRIVATE_MAGIC))
		if PRIVATE_MAGIC == magic:
			version = 1
			(keyname, key) = gale_pack.pop_nulltermstr(key)
		elif PRIVATE_MAGIC2 == magic:
			version = 2
			(keyname, key) = gale_pack.pop_lenstr(key, chars=1)
		else:
			pygale.call_error_handler('invalid private key format')
			return None
		if DEBUG: print 'Found private key with name:', keyname

		privkey = rsaref.R_RSA_PRIVATE_KEY()
		(bits, key) = gale_pack.pop_int(key)
		if bits < rsaref.MIN_RSA_MODULUS_BITS or \
			bits > rsaref.MAX_RSA_MODULUS_BITS:
			pygale.call_error_handler('wrong number of bits in private key')
			return None
		privkey.bits = bits
		(modulus, key) = gale_pack.pop_rle(key, rsaref.MAX_RSA_MODULUS_LEN)
		privkey.modulus = modulus
		(pubexp, key) = gale_pack.pop_rle(key, rsaref.MAX_RSA_MODULUS_LEN)
		privkey.publicExponent = pubexp
		(exp, key) = gale_pack.pop_rle(key, rsaref.MAX_RSA_MODULUS_LEN)
		privkey.exponent = exp
		(prime, key) = gale_pack.pop_rle(key, rsaref.MAX_RSA_PRIME_LEN*2)
		privkey.prime = prime
		(primeexp, key) = gale_pack.pop_rle(key, rsaref.MAX_RSA_PRIME_LEN*2)
		privkey.primeExponent = primeexp
		(coef, key) = gale_pack.pop_rle(key, rsaref.MAX_RSA_PRIME_LEN)
		privkey.coefficient = coef

		if not self.__keys.has_key(keyname):
			keyobj = sign.Key(keyname)
			keyobj.setprivate(privkey)
			self.__keys[keyname] = keyobj
		else:
			keyobj = self.__keys[keyname]
			keyobj.setprivate(privkey)

		return self.__keys[keyname]

	# Cache a public key if one can be found in the key path
	# Return the name of the key if found, None otherwise
	def find_pubkey_fromdisk(self, keyobj, callback):
		name = keyobj.name()
		if DEBUG > 2: print 'f_pk_fd: %s' % (name)
		for (prefix, suffix, trust) in self.KEYPATH:
			path = os.path.join(prefix, suffix, name)
			if DEBUG: print 'authcache: checking', path
			if os.path.exists(path):
				if DEBUG: print 'found it'
				f = open(path, 'rb')
				keydata = f.read()
				f.close()
				keyobj.setdata(keydata)
				sign.import_pubkey(keyobj,
					lambda k, cb=callback, s=self:
					s.cache_key_cb(k, cb), trust)
				return
		# Otherwise, we didn't find it; return None
		keyobj.setpublic(None)
		callback(keyobj)
		return

	def cache_key_cb(self, keyobj, callback):
		# Save it to in-memory cache
		if keyobj.public():
			if DEBUG: print 'caching key:', keyobj
			self.__keys[keyobj.name()] = keyobj
		callback(keyobj)

	# AKD
	def find_pubkey_akd(self, keyobj, callback):
		name = keyobj.name()
		if DEBUG: print 'Finding public key', name
		if DEBUG: print 'Callback:', callback
		if '@' in name:
			i = string.find(name, '@')
			domain = name[i+1:]
			user = name[:i]
		else:
			# Hmm, we can't request a domain key without the user
			# Invalid key name
			keyobj.setpublic(None)
			callback(keyobj)
			return

		# First see when was the last time we did AKD for this key
		if self.__last_akd.has_key(name):
			last_checked = self.__last_akd[name]
			now = time.time()
			if now - last_checked < AKD_INTERVAL:
				# Chain requests
				if DEBUG: print 'not doing AKD; done it too recently'
				pygale.call_update_handler('Chaining to existing key '+
					'request for %s' % name)
				self.__keychain[name] = self.__keychain.get(name, []) +\
					[callback]
				return
		else:
			self.__last_akd[name] = time.time()
	
		self.__keychain[name] = self.__keychain.get(name, []) +\
			[callback]
		if DEBUG: print 'key %s not found in cache; trying AKD' % name
		pygale.call_update_handler('Requesting key %s' % name)
		client = pygale.GaleClient()
		client.connect(lambda h, s=self, d=domain, u=user, n=name, k=keyobj,
			c=client: s.find_pubkey_akd2(h, d, u, n, c, k))

	def find_pubkey_akd2(self, hostname, domain, user, name, client,
		keyobj):
		if hostname is None:
			pygale.call_error_handler('Unable to connect to gale ' +
				'server for AKD')
			keyobj.setpublic(None)
			self.cleanup_akd(keyobj)
			return

		client.sub_to(KEY_RESPONSE_CAT % (domain, user))
		p = pygale.Puff()
		p.set_cat(KEY_REQUEST_CAT % (domain, user))
		p.set_text('question/key', name)
		client.transmit_puff(p)
		timeout_callback = lambda k=keyobj, s=self, l=client:\
			s.handle_AKD_timeout(k, l)
		client.set_puff_callback(lambda p, s=self, k=keyobj,
			l=client, t=timeout_callback, d=domain, i=name:
			s.handle_key_puff(p, k, l, t, d, i))
		engine.timeout.add_callback(AKD_TIMEOUT, timeout_callback)
	
	def cleanup_akd(self, keyobj):
		name = keyobj.name()
		if keyobj.public():
			for keyname in [name] + keyobj.get_links_to_me():
				self.__keys[keyname] = keyobj
		if DEBUG:
			print 'Cleaning up after', name
			print 'Links to me:', keyobj.get_links_to_me()

		for name in [keyobj.name()] + keyobj.get_links_to_me():
			if self.__keychain.has_key(name):
				# make copy of callbacks
				cbs = self.__keychain[name][:]
				del self.__keychain[name]
				for cb in cbs:
					if DEBUG: print 'Cleanup CB:', cb
					cb(keyobj)
			if self.__last_akd.has_key(name):
				del self.__last_akd[name]

	def handle_AKD_timeout(self, key, client):
		if DEBUG: print 'AKD request timeout.'
		pygale.call_update_handler('key request for %s timed out' %
			key.name())
		client.del_puff_callback()
		key.setdata(None)
		self.cleanup_akd(key)

	def handle_key_puff(self, puff, keyobj, client, timeout_callback,
		domain, id):
		keyobj.setpublic(None)
		if DEBUG: print 'Key request puff returned'
		if puff.get('answer/key/error', None) is not None:
			client.del_puff_callback()
			engine.timeout.del_callback(timeout_callback)
			signer = puff.get_signer(None)
			if signer is None or signer != domain:
				if DEBUG: print 'Received unverified AKD response'
				pygale.call_update_handler(
					'Received unverified AKD response from ' + `signer`)
				pygale.call_update_handler(
					'Unverified response: ' + `puff.get(
						'answer/key/error')`)
			pygale.call_update_handler(puff.get('answer/key/error'))
			self.cleanup_akd(keyobj)
			return
		# TODO: check that this key actually works before cancelling the
		# callbacks
		if puff.get('answer/key', None) is not None:
			key = puff.get('answer/key')
			keyobj.setdata(key)
			if DEBUG: print 'Importing pubkey bits'
			sign.import_pubkey(keyobj, lambda k, to=timeout_callback,
				c=client, s=self: s.finish_key_puff(k, c, to), 0)
	
	def finish_key_puff(self, keyobj, client, timeout_callback):
		if not keyobj.public():
			# Got a bad public key; don't bother cleaning up just yet
			if DEBUG: print 'Got a bad public key; waiting for another'
			return
		if DEBUG: print 'Finished handling key puff; cleaning up'
		client.del_puff_callback()
		engine.timeout.del_callback(timeout_callback)
		self.cleanup_akd(keyobj)

	# Find a private key on disk
	# Return the key object, or None if not found
	def find_privkey_fromdisk(self, name):
		for (prefix, suffix) in self.PRIVKEYPATH:
			path = os.path.join(prefix, suffix, name)
			#print 'authcache: checking', path
			if os.path.exists(path):
				f = open(path, 'rb')
				key = f.read()
				f.close()
				return self.import_privkey(key)

		return None

	def save_pubkey_todisk(self, keyobj):
		if keyobj.data() is None:
			if DEBUG: print 'Got empty public key to save to disk!'
			return
#		if DEBUG: print 'save_pubkey_todisk(%s)' % keyobj.name()
		# Save it to a file
		tempfile.tempdir = self.CACHE_PATH
		fname = tempfile.mktemp()
		keypath = os.path.join(self.CACHE_PATH, keyobj.name())
		if os.path.exists(keypath):
			pygale.call_update_handler('overwriting old '+
				'public key %s with a new one' % keyobj.name())
		try:
			f = open(fname, 'wb')
			f.write(keyobj.data())
			f.close()
			# Fix for win32: rename won't overwrite an existing file
			if sys.platform == 'win32':
				if os.path.exists(keypath):
					os.unlink(keypath)
			os.rename(fname, keypath)
		except IOError, e:
			pygale.call_update_handler(
				'Error saving key to cache: ' + str(e))
	
	def find_pubkey_list(self, keyobjs, callback):
		c = Collector(map(lambda x: x.name(), keyobjs))
		for key in keyobjs:
			self.find_pubkey(key, lambda key, c=c, cb=callback, s=self:
				s.collector_callback(key, c, cb))
	
	def collector_callback(self, key, collector, callback):
		collector.setoutput(key.name(), key)
		if collector.done():
			callback(collector.getvalues())
	
	def find_pubkey(self, keyobj, callback):
#		if DEBUG: print 'Find_pubkey: %s (%s)' % (`keyobj`, keyobj.name())
		if DEBUG: print 'Find_pubkey (%s), cb:' % keyobj.name(), callback
		keyobj.setpublic(None)
		if self.__keys.has_key(keyobj.name()) and\
			self.__keys[keyobj.name()].public():
			if DEBUG: print 'Returning key from in-memory cache'
			cached_key = self.__keys[keyobj.name()]
			callback(cached_key)
			return

		# Go to the disk cache
		self.find_pubkey_fromdisk(keyobj, lambda k, s=self, cb=callback:
			s.find_pubkey2(k, cb))
	
	def find_pubkey2(self, keyobj, callback):
		if keyobj.public() is not None:
			# Put it in the in-memory cache
			self.__keys[keyobj.name()] = keyobj
			callback(keyobj)
			return
		
		# Oops, we gotta do AKD
		self.find_pubkey_akd(keyobj, lambda k, s=self, cb=callback:
			s.find_pubkey3(k, cb))
	
	def find_pubkey3(self, key, callback):
		if key.public() is not None:
			# TODO: cache key here
			self.__keys[key.name()] = key
			callback(key)
			return
#		pygale.call_error_handler('unable to find public key %s' %
#			key.name())
		if DEBUG: print 'In find_pubkey3, calling callback'
		if DEBUG: print 'callback:', callback
		callback(key)

def save_pubkey_todisk(*args):
	return apply(KeyCache.save_pubkey_todisk, args)
def find_pubkey_fromdisk(*args):
	return apply(KeyCache.find_pubkey_fromdisk, args)
def find_privkey_fromdisk(*args):
	return apply(KeyCache.find_privkey_fromdisk, args)

def find_pubkey_akd(*args):
	return apply(KeyCache.find_pubkey_akd, args)

def find_pubkey(*args):
	return apply(KeyCache.find_pubkey, args)
def find_pubkey_list(*args):
	return apply(KeyCache.find_pubkey_list, args)
def get_key(*args):
	return apply(KeyCache.get_key, args)

def init():
	global KeyCache
	if DEBUG: print 'Initializing keycache'
	KeyCache = KeyCacheClass()

	if not gale_env.has_key('GALE_SYS_DIR'):
		print 'Warning: GALE_SYS_DIR not set; using ""'
	GALE_SYS_DIR = gale_env.get('GALE_SYS_DIR', '')
	KeyCache.KEYPATH = [
		(HOME, os.path.join('auth','trusted'), TRUSTED),
		(HOME, os.path.join('auth','local'), NORMAL),
		(HOME, os.path.join('auth','cache'), NORMAL),
		(GALE_SYS_DIR, os.path.join('auth', 'trusted'), TRUSTED),
		(GALE_SYS_DIR, os.path.join('auth', 'local'), NORMAL),
		(GALE_SYS_DIR, os.path.join('auth', 'cache'), NORMAL)]
	KeyCache.CACHE_PATH = os.path.join(GALE_SYS_DIR, 'auth', 'cache')

	KeyCache.PRIVKEYPATH = [
		(HOME, os.path.join('auth','private')),
		(GALE_SYS_DIR, os.path.join('auth', 'private'))]

	# Save reference to rsarefc destructor; we need to call it
	global KeyDestructor
	KeyDestructor = [rsaref.rsarefc.delete_R_RSA_PUBLIC_KEY,
		rsaref.rsarefc.delete_R_RSA_PUBLIC_KEY]

def shutdown():
	global KeyCache
	del KeyCache
