#!/usr/bin/env python

import getopt, sys, os, select, errno, re
import socket, pickle, string, time, traceback, whrandom
import rsaref
import engine, prettypuff, authcache, gale_pack, sign, gale_env
from types import *

# Enable debugging output
DEBUG = 0

PyGaleErr = 'PyGale Error'
PyGaleWarn = 'PyGale Warning'

DEFAULT_GALE_PORT = 11512

# Fragment types
FRAG_TYPE_TEXT = 0
FRAG_TYPE_BINARY = 1
FRAG_TYPE_TIME = 2
FRAG_TYPE_INT = 3
FRAG_TYPE_LIST = 4
# Gale constants
# TODO: check to see which are actually needed in this file
SIG_MAGIC = 'h\023\001\000'
PRIVATE_MAGIC = 'h\023\000\001'
KEY_MAGIC = 'h\023\000\000'
KEY_MAGIC2 = 'h\023\000\002'
PRIVATE_MAGIC2 = 'h\023\000\003'
CIPHER_MAGIC = 'h\023\002\000'
CIPHER_MAGIC2 = 'h\023\002\001'
# Error handling function
ERROR_HANDLER = None
# Update handling function
UPDATE_HANDLER = None
# Maximum time (in seconds) before next retry attempt
MAX_RETRY_DELAY = 120

def gale_u32_size():
	return 4

def gale_wch_size():
	return 2


class Blocker:
	wait = 1
	val = None
	def done(self):
		self.wait = 0
	def done_setval(self, v):
		self.wait = 0
		self.val = v

class Puff:
	def __init__(self, dict=None):
		self.__fragments = {}
		if dict is not None:
			self.__fragments.update(dict)
		self.__cat = None
		self.__signer = None
		self.__recipients = []
	
	def set_cat(self, cat):
		self.__cat = cat
	
	def get_cat(self, default=''):
		if self.__cat is None:
			return default
		else:
			return self.__cat
	
	def get_signer(self, default=''):
		if self.__signer is None:
			return default
		else:
			return self.__signer

	def get_recipients(self):
		return self.__recipients

	def set_recipients(self, ids):
		self.__recipients = ids
	def is_encrypted(self):
		# TODO: remove this once no one has pre-0.9.7d puffs lying around
		# In 0.9.7d I added self.__encrypted to each puff (boolean)
		# In 0.9.8c I removed the self.__encrypted variable (but I need
		#    it for old puffs); use self.__recipients instead
		try:
			ret = self.__encrypted
		except AttributeError:
			ret = 0
		return ret

	def get(self, fragname, default=None):
		if self.__fragments.has_key(fragname):
			return self.__fragments[fragname][1]
		else:
			return default
	
	def set_text(self, name, val):
		self.__fragments[name] = (FRAG_TYPE_TEXT, val)
	
	def set_time(self, name, val):
		self.__fragments[name] = (FRAG_TYPE_TIME, val)

	def set_int(self, name, val):
		self.__fragments[name] = (FRAG_TYPE_INT, val)
	
	def set_binary(self, name, val):
		self.__fragments[name] = (FRAG_TYPE_BINARY, val)
	
	def fragments(self):
		return self.__fragments
	
	def __setitem__(self, name, value):
		self.__fragments[name] = value
	
	def __repr__(self):
		return ('[%s] ' % self.__cat) + repr(self.__fragments)
	
	def swizzled(self):
		# Construct puff bits
		puffbits = gale_pack.push_lenstr(self.get_cat(''), chars=0) +\
			gale_pack.push_int(0)
		if DEBUG >= 2: print 'Puff with category:', `puffbits`
		puffbits = puffbits + gale_pack.dict_to_group(self.fragments())
		if DEBUG >= 2: print 'With fragments:', `puffbits`
		if DEBUG >= 2: print 'Puff datalen is', len(puffbits)
		return puffbits
	
	# Consruct a signed message fragment
	# signer is the name of the signing key, e.g. "egnor@ofb.net"
	def sign_message(self, signer):
		p = Puff()
		p.__cat = self.__cat
		p.__signer = self.__signer = signer
		p.__recipients = self.__recipients

		# TODO: use the equivalent of gale_user() here
		if DEBUG: print 'Signing message with signer', signer
		input_data = ''
		input_data = input_data + gale_pack.push_int(0)
		input_data = input_data + gale_pack.dict_to_group(self.__fragments)

		privkey = authcache.get_key(signer)
		if not (privkey and privkey.private()):
			# If the private key isn't cached, try to find it
			privkey = authcache.find_privkey_fromdisk(signer)
			if not (privkey and privkey.private()):
				call_error_handler('No private key %s to sign with' %
					signer)
				return None

		context = rsaref.R_SIGNATURE_CTX()
		context.R_SignInit(rsaref.DA_MD5)
		context.R_SignUpdate(input_data, len(input_data))
		buf = rsaref.Buffer()
		buf.create(rsaref.MAX_SIGNATURE_LEN)
		(retval, siglen) = context.R_SignFinal(buf, privkey.private())
		if DEBUG: print 'SignFinal returned:', retval
		if retval:
			call_error_handler('Failure signing message with key ' + signer)
			return None

		# pack a new fragment
		sig = ''
		sig = SIG_MAGIC + gale_pack.push_int(siglen) + buf.data[:siglen] +\
			make_stub_pub(signer)
		out = gale_pack.push_int(len(sig)) + sig + input_data

		p.__fragments = {'security/signature':
			(FRAG_TYPE_BINARY, out)
		}

		return p
	
	def encrypt_message(self, recipient_list, callback):
		if type(recipient_list) is not ListType:
			recipient_list = [recipient_list]

		if DEBUG: print 'Encrypting message to', recipient_list
		keyobjlist = map(sign.Key, recipient_list)
		if DEBUG: print 'Encryption keys:', keyobjlist
		try:
			authcache.find_pubkey_list(keyobjlist,
				lambda l, cb=callback, s=self: s.encrypt_message2(l, cb))
		except PyGaleErr, e:
			call_error_handler(e)
			callback(None)
	
	def encrypt_message2(self, recipient_keys, callback):
		keynames = map(lambda x: x.name(), recipient_keys)
		if DEBUG: print 'encrypt_message2: keys', keynames
		for key in recipient_keys:
			if DEBUG: print 'em2: processing key', key
			if key.public() is None:
				if DEBUG: print 'em2: no key', key
				call_error_handler('no public key (%s) to encrypt with' %\
					key.name())
				callback(None)
				if DEBUG: print 'em2: returning'
				return

		# Now do the sealing
		if DEBUG: print 'encrypt_message2 context'
		context = rsaref.R_ENVELOPE_CTX()
		rnd = rsaref.R_RANDOM_STRUCT()
		num = len(recipient_keys)
		keys = map(lambda x: x.public(), recipient_keys)

		rand = initrand()
		(retval, ekeys, ekeylen, iv) = context.R_SealInit(num,keys,
			rsaref.EA_DES_EDE3_CBC,rand)

		if DEBUG: print 'encrypt_message2 top'
		plain_data = gale_pack.push_int(0) + gale_pack.dict_to_group(
			self.__fragments)
		n = len(CIPHER_MAGIC2) + len(iv) + gale_u32_size() + len(plain_data)
		for i in range(num):
			n = n + (gale_u32_size() + (len(recipient_keys[i].name()) *
				gale_wch_size()))
			n = n + ekeylen[i] + gale_u32_size()

		cipher = CIPHER_MAGIC2 + iv + gale_pack.push_int(num)
		for i in range(num):
			cipher = cipher + gale_pack.push_lenstr(recipient_keys[i].name())
			cipher = cipher + gale_pack.push_int(ekeylen[i])
			cipher = cipher + ekeys[i]

		if DEBUG: print 'encrypt_message2 middle'
		buf = rsaref.Buffer()
		buf.create(len(plain_data) + 7)
		(retval, updatelen) = context.R_SealUpdate(buf, plain_data,
			len(plain_data))
		if retval:
			call_error_handler('R_SealUpdate returned %i' % retval)
			callback(None)
			return
		cipher = cipher + buf.data[:updatelen]
		buf = rsaref.Buffer()
		buf.create(8)
		(retval, finallen) = context.R_SealFinal(buf)
		if retval:
			call_error_handler('R_SealFinal returned %i' % retval)
			callback(None)
			return
		cipher = cipher + buf.data[:finallen]

		if DEBUG: print 'encrypt_message2 constructing new puff'
		p = Puff()
		p.__cat = self.__cat
		p.__signer = self.__signer
		p.__recipients = self.__recipients = map(lambda x: x.name(),
			recipient_keys)
		p.__fragments = {'security/encryption':
			(FRAG_TYPE_BINARY, cipher)}
		if DEBUG: print 'calling callback from encrypt_message2'
		callback(p)

	def verify(self, callback):
		if not self.__fragments.has_key('security/signature'):
			self.__signer = '*unsigned*'
			callback(self)
			return self

		p = Puff()
		p.__cat = self.__cat
		p.__signer = None
		p.__recipients = self.__recipients

		frag = self.__fragments['security/signature'][1]
		(sig_len, frag) = gale_pack.pop_int(frag)
		(sig, frag) = gale_pack.pop_data(frag, sig_len)
		p.__fragments.update(gale_pack.group_to_dict(frag))
		sign.decode_sig(sig, frag, lambda key, s=self, p=p, c=callback:
			s.decode_done(key, p, c))
		return p

	def decode_done(self, key, puff, callback):
		if key.public() is None:
			signer = '*unverified*'
		else:
			signer = key.name()
		if DEBUG: 'Verify done: %s' % (signer)
		puff.__signer = signer
		callback(puff)
	
	def decrypt(self):
		if not self.__fragments.has_key('security/encryption'):
			self.__recipients = []
			return self

		p = Puff()
		p.__cat = self.__cat
		p.__signer = self.__signer
		p.__recipients = self.__recipients

		frag = self.__fragments['security/encryption'][1]
		encr_id, plain_msg, frag = self.__decrypt_msg(frag)
		p.__recipients = encr_id
		p.__fragments.update(gale_pack.group_to_dict(plain_msg))
		return p

	def __decrypt_msg(self, cipher):
		if DEBUG: print 'Decrypting message...'
		# init variables
		plain_msg = None

		# auth_decrypt
		(magic, cipher) = gale_pack.pop_data(cipher, len(CIPHER_MAGIC))
		if DEBUG: print 'Magic is:', `magic`
		if magic == CIPHER_MAGIC:
			version = 1
		elif magic == CIPHER_MAGIC2:
			version = 2
		else:
			raise PyGaleWarn, 'Invalid encryption magic'

		# constant length 8 byte "iv" data
		(iv, cipher) = gale_pack.pop_data(cipher, 8)
		(keycount, cipher) = gale_pack.pop_int(cipher)
		if keycount != 1:
			if DEBUG: print 'Found keycount:', keycount

		found_key = 0
		ekeyname = []
		for i in range(keycount):
			if version == 1:
				(keyname, cipher) = gale_pack.pop_nulltermstr(cipher)
			else:
				(keyname, cipher) = gale_pack.pop_lenstr(cipher, chars=1)
			(keylen, cipher) = gale_pack.pop_int(cipher)
			(keydata, cipher) = gale_pack.pop_data(cipher, keylen)
			if DEBUG: print 'Found key name:', keyname
			ekeyname.append(keyname)
			if keylen > rsaref.MAX_ENCRYPTED_KEY_LEN or\
				len(keydata) < keylen:
				raise PyGaleWarn, 'Invalid encrypted key length'
			# If we already found a working private key, no need to find
			# another one
			if found_key: continue

			private_key = authcache.get_key(keyname)
			if not (private_key and private_key.private()):
				# We don't have the private key cached;
				# try to import it from disk
				private_key = authcache.find_privkey_fromdisk(keyname)

			if (private_key and private_key.private()):
				# Now, if we have a private key, we can decrypt
				ekey = keydata
				ekeylen = keylen
				
				# Test if this key works
				env = rsaref.R_ENVELOPE_CTX()
				ret = env.R_OpenInit(rsaref.EA_DES_EDE3_CBC, ekey,
					ekeylen, iv, private_key.private())
				if not ret:
					# Success
					found_key = 1

		if not found_key:
			# If we haven't found any private key that works
			# No private key found to decrypt this message
			raise PyGaleWarn, "No private key(s) to decrypt puff on category %s" % self.get_cat()

		buf = rsaref.Buffer()
		buf.create(len(cipher) + 7)
		ret, len_plain = env.R_OpenUpdate(buf, cipher, len(cipher))
		if ret:
			raise PyGaleWarn, 'R_OpenUpdate returned %i' % ret
		plain_msg = buf.data[:len_plain]
		buf = rsaref.Buffer()
		buf.create(7)
		ret, len_plain = env.R_OpenFinal(buf)
		if ret:
			raise PyGaleWarn, 'R_OpenFinal returned %i' % ret
		plain_msg = plain_msg + buf.data[:len_plain]

		if DEBUG: print 'Decryption succeeded, encrypted to:', ekeyname
		return (ekeyname, plain_msg, '')

def initrand():
	r = rsaref.R_RANDOM_STRUCT()
	r.R_RandomInit()
	while 1:
		(retval, needed) = r.R_GetRandomBytesNeeded()
		if not needed: break
		randbytes = getrandom()
		r.R_RandomUpdate(randbytes, len(randbytes))
	return r

def getrandom():
	import os, whrandom
	ret = ''
	ret = ret + gale_pack.push_int(os.getpid()) + gale_pack.push_int(int(time.time()))
	whrandom.seed(int(time.time()) % 256)
	for i in range(5):
		ret = ret + gale_pack.push_int(whrandom.randint(0, 2147483646))
	md5 = rsaref.MD5_CTX()
	md5.MD5Init()
	md5.MD5Update(ret, len(ret))
	return md5.MD5Final()

def make_stub_pub(signer):
	return KEY_MAGIC2 + gale_pack.push_lenstr(signer)

class GaleClient:
	def __init__(self, hostname=None, eng=None, retry=1):
		if hostname is None:
			# Default connections
			proxy = gale_env.get('GALE_PROXY', None)
			if proxy is not None:
				self._hosts = re.split('[, ]+', proxy)
			else:
				domain = gale_env.get('GALE_DOMAIN', None)
				if domain is None:
					print 'GaleClient initialization error: unable ' +\
						'to determine GALE_DOMAIN'
					self._hosts = []
				else:
					self._hosts = [domain, 'gale.' + domain,
						domain + '.gale.org']
		else:
			self._hosts = [hostname]
		self._retry = retry		# whether to automatically reconnect
		self._sock = None
		self._buffer = ''
		self._puff_callback = None
		self._verify_callback = None
		self._on_connect = None
		self._on_disconnect = None
		self._puffs = []
		if eng == None:
			self._engine = engine.engine
		else:
			self._engine = eng
		self.__sockets_in_progress = []
		self.__connected = 0
		self._retry_delay = 1	# how many seconds before this retry

	def __del__(self):
		if DEBUG: print 'GaleClient.__del__: Client dying'
		if self._puff_callback:
			self._engine.del_callback(self._sock)

	def set_onconnect(self, cb):
		self._on_connect = cb
	
	def set_ondisconnect(self, cb):
		self._on_disconnect = cb
	
	def socket(self):
		return self._sock
	
	# The callback will be called with a hostname, or with None if connect
	# failed
	def connect(self, callback = None):
		if DEBUG: print 'starting connect'
		if callback is None:
			blocker = Blocker()
			cb = lambda h, b=blocker: b.done_setval(h)
		else:
			cb = callback

		self.__connected = 0
		self.__sockets_in_progress = []
		for hostport in self._hosts:
			if ':' in hostport:
				index = string.index(hostport, ':')
				portstr = hostport[index+1:]
				try:
					port = int(portstr)
				except ValueError:
					call_error_handler(
						'Error parsing server string: %s' % hostport)
					port = DEFAULT_GALE_PORT
				host = hostport[:index]
			else:
				host = hostport
				port = DEFAULT_GALE_PORT
			sock = socket.socket(socket.AF_INET,
				socket.SOCK_STREAM)
			sock.setblocking(0)
			sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
			self.__sockets_in_progress.append(sock)
			self._engine.add_write_callback(sock, lambda self=self,
				s=sock, h=hostport, cb=cb: self.connect2(s, h, cb))
			if DEBUG: print 'Connecting to %s' % hostport
			try:
				sock.connect((host, port))
			except socket.error, e:
				if e[0] == errno.EINPROGRESS or e[0] == 10035:
					if DEBUG: print 'Connect in progress'
					continue
				else:
					if DEBUG: print 'Connect failed to', host
					self._engine.del_write_callback(sock)
					self.__sockets_in_progress.remove(sock)
					continue
			# Connected immediately
			self.connect2(sock, hostport, cb)

		# Check once after all hosts have been processed
		if not self.__sockets_in_progress:
			cb(None)
			return

		if callback is None:
			while blocker.wait:
				engine.engine.process()
			return blocker.val
	
	def connect2(self, sock, host, callback):
		if DEBUG: print 'in connect2'
		self._engine.del_write_callback(sock)
		if self.__connected:
			# Someone else already connected
			if DEBUG: print 'already connected to a server'
			self.__sockets_in_progress.remove(sock)
			return

		# I suspect this block isn't necessary
		r, w, e = select.select([], [sock], [], 0)
		if sock not in w:
			# Something weird's going on
			call_error_handler('Weirdness connecting to %s' % host)
			self.__sockets_in_progress.remove(sock)
			if not self.__sockets_in_progress:
				call_error_handler(
					'Error: unable to connect to any server')
				callback(None)
			return

		# Check for error
		if DEBUG: print 'checking socket options'
		if sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR) != 0:
			# unsuccessful
			if DEBUG: print 'socket options indicate error connecting' +\
				' to ' + host
			self.__sockets_in_progress.remove(sock)
			if not self.__sockets_in_progress:
				call_update_handler(
					'Error: unable to connect to any server')
				callback(None)
			return

		# Otherwise, we're connected
		if DEBUG: print 'it seems we are connected'
		try:
			# Send our version string
			sock.send(gale_pack.pack32bit(1))
		except socket.error, e:
			call_error_handler('Error sending version string to %s: %s'
				% (host, str(e)))
			self.__sockets_in_progress.remove(sock)
			if not self.__sockets_in_progress:
				callback(None)
			return

		self._engine.add_callback(sock, lambda self=self, s=sock, h=host,
			cb=callback: self.connect3(s, h, '', cb))

	# Hmm, shouldn't this possibly fail in some way?
	def connect3(self, sock, host, buf, callback):
		self._engine.del_callback(sock)
		if self.__connected:
			self.__sockets_in_progress.remove(sock)
			return
		try:
			r = sock.recv(1024)
		except socket.error:
			# It's possible to get a "connection reset" error here
			# Can't connect to server
			self.__sockets_in_progress.remove(sock)
			return

		r = buf + r
		if len(r) >= 4:
			# We got the handshake
			server_version = gale_pack.unpack32bit(r[:4])
			if DEBUG: print 'Server version is', server_version
			self._buffer = r[4:]
			self.__connected = 1
			self._sock = sock
			callback(host)
			if self._on_connect is not None:
				self._on_connect(host)
		else:
			self._engine.add_callback(sock, lambda self=self, s=sock,
				h=host, b=r, cb=callback: self.connect3(s, h, b, cb))
	
	
	# Attempt to reconnect to server
	def retry(self):
		self._retry_delay = 1
		self.continue_retry()
	
	def continue_retry(self):
		call_update_handler('Retrying server connection')
		self.connect(self.retry_done)

	def retry_done(self, host):
		if DEBUG: print 'In retry_done, host:', host
		if host is None:
			# Retry failed
			if self._retry_delay == 1:
				wait = 1
			else:
				wait = whrandom.randint(1, self._retry_delay)
			self._retry_delay = self._retry_delay + wait
			if self._retry_delay > MAX_RETRY_DELAY:
				self._retry_delay = self._retry_delay / 2
			call_error_handler('Retry failed; waiting %i seconds' %
				self._retry_delay)
			self._engine.add_timeout(self._retry_delay * 1000,
				self.continue_retry)
			return
		call_update_handler('Reconnected to server at %s' % host)
		self.sub_to(self._category)
		call_update_handler('Subscribed to %s' % self._category)
		# If retry succeeded, we don't need to do anything

	# Subscribe to a category.  If category is None, subscribe to nothing
	# ("-").
	def sub_to(self, category):
		if category is None:
			# Subscribe to nothing
			category = '-'

		self._category = category
		req = gale_pack.pack32bit(2)
		ucat = gale_pack.push_string(category)
		req = req + gale_pack.pack32bit(len(ucat))
		req = req + ucat
		if self._sock is not None:
			self._sock.send(req)
		else:
			print 'Unable to subscribe; no server connection!'
		if DEBUG: print 'Subscribed to:', category
	
	def transmit_puff(self, p, will=0):
		if will:
			opcode = gale_pack.push_int(1)
		else:
			opcode = gale_pack.push_int(0)
		puffbits = p.swizzled()
		# construct server request
		req = opcode + gale_pack.push_int(len(puffbits)) + puffbits
#		if DEBUG: print 'Sending puff:', `req`
		if self._sock is None:
			raise PyGaleErr, 'Error sending puff: dead socket!'
		try:
			self._sock.send(req)
		except socket.error, e:
			if e[0] == errno.EPIPE:
				raise PyGaleErr, 'Error sending puff: connection closed'
			else:
				raise PyGaleErr, 'Error sending puff: %s' % e[1]


	def set_puff_callback(self, callback):
		self._puff_callback = callback
		self._engine.add_callback(self._sock, self.handle_read)


	def set_verify_callback(self, callback):
		self._verify_callback = callback


	def del_puff_callback(self):
		if self._puff_callback:
			self._engine.del_callback(self._sock)
			self._puff_callback = None

	def disconnect(self):
		self.del_puff_callback()
		self._sock.close()
		if self._on_disconnect is not None:
			self._on_disconnect()

	def handle_read(self):
		try:
			r = self._sock.recv(1024)
		except socket.error, e:
			self.del_puff_callback()
			call_error_handler('Server died: %s' % str(e))
			if self._on_disconnect is not None:
				self._on_disconnect()
			# Wait half a second, then retry
			self._engine.add_timeout(500, self.retry)
			return None
		if r == '':
			self.del_puff_callback()
			call_error_handler('Server closed connection (read)')
			if self._on_disconnect is not None:
				self._on_disconnect()
			# Wait half a second, then retry
			self._engine.add_timeout(500, self.retry)
			return None
		self._buffer = self._buffer + r
		while len(self._buffer) > 8:
			opcode = gale_pack.unpack32bit(self._buffer[0:4])
			datalen = gale_pack.unpack32bit(self._buffer[4:8])
			if len(self._buffer) - 8 < datalen:
				break

			self._buffer = self._buffer[8:]
			puffdata = self._buffer[:datalen]
			self._buffer = self._buffer[datalen:]
			if DEBUG >= 2: print 'Processing puff len', datalen
			if opcode == 0:
				try:
					self._puffs.append(self.process_puff(puffdata))
				except PyGaleWarn, e:
					if ERROR_HANDLER:
						ERROR_HANDLER(e)
			elif opcode == 1:
				# We should never get wills
				print "Wills not handled"
			elif opcode == 2:
				print "Gimme not handled"


	# Non-blocking call to get next puffs queued
	def next_puffs(self):
		puffs = []
		self.set_puff_callback(lambda p, puffs=puffs: puffs.append(p))
		self._engine.process(0)
		return puffs

	# Read from socket forever, calling handler(Puff: p) whenever we
	# receive a complete puff
	# This will only terminate when the server closes the connection
	def next(self, handler):
		self.set_puff_callback(handler)
		while 1:
			self._engine.process()

	def process_puff(self, p):
		if DEBUG >= 2: print 'Processing puff:', `p`
		(cat, p) = gale_pack.pop_lenstr(p)
		if DEBUG: print 'Puff category:', cat

		# Fragment list
		p = Puff(gale_pack.group_to_dict(p))
		p.set_cat(cat)
		if DEBUG >= 2: print 'Fragments:', p.fragments()
		p = p.decrypt()
		if self._verify_callback:
			p = p.verify(lambda p, s=self: s._verify_callback(p))
		elif self._puff_callback:
			p = p.verify(lambda p, s=self: s._puff_callback(p))
		else:
			print 'Yikes! no verify or puff callback registered!'
		# If you don't tell us you want to be notified of verification
		# we will not call your puff callback until verification is done.
		# (tlau) There seems like there could be a race condition here if
		# self._verify_callback is changed while the puff is being
		# verified.  Perhaps the verify callback (above) ought to take
		# one function if verify_callback is now set, and a different one
		# if it isn't?  That way the callbacks are called as is
		# appropriate for the setting at the time the puff arrives.
		# In particular, if self._verify_callback is removed during the
		# verify operation, it could result in puff_callback being called
		# twice on the same puff.
		#p = p.verify(lambda p, s=self: s.verify_done(p))
		# I think I fixed this above, though it ought to be tested.
		if self._verify_callback:
			if self._puff_callback:
				self._puff_callback(p)
			else:
				print 'Yikes! no puff callback registered!'

#	def verify_done(self, puff):
#		if self._verify_callback:
#			self._verify_callback(puff)	
#		else:
#			self._puff_callback(puff)

# ----------------------------------------------------------------------
# Misc utility functions

def expand_id(name):
	if '@' in name:
		return name
	else:
		return '%s@%s' % (name, gale_env.get('GALE_DOMAIN',
			'GALE_DOMAIN_UNSET'))

# Calls callback with comment_string
def id_comment(name, callback = None):
	if callback is None:
		blocker = Blocker()
		cb = lambda k, blocker=blocker: blocker.done_setval(k)
	else:
		cb = callback
	keyobj = sign.Key(name)
	try:
		authcache.find_pubkey(keyobj, lambda k, c=cb: id_comment_done(k, c))
	except PyGaleErr, e:
		call_error_handler(e)
		return
	if not callback:
		while blocker.wait:
			engine.engine.process(authcache.AKD_TIMEOUT)
		return blocker.val

def id_comment_done(key, callback):
	callback(key.comment())

def gale_user():
	# Find the current user
	gdomain = gale_env.get('GALE_DOMAIN', 'GALE_DOMAIN_UNSET')
	guser = gale_env.get('GALE_USER', os.environ.get('USER', 'nobody'))
	return '%s@%s' % (guser, gdomain)

def id_category(name, midstr='', postfix=''):
	i = string.find(name, '@')
	username = name[:i]
	domain = name[i+1:]
	return '@%s/%s/%s/%s' % (domain, midstr, username, postfix)

def dom_category(domain, postfix=''):
	return '@%s/%s' % (domain, postfix)

def gale_domain():
	return gale_env.get('GALE_DOMAIN', 'GALE_DOMAIN_UNSET')

# Calls callback on (gale_id, comment)
def lookup_id(name, callback = None):
	if not '@' in name:
		gdomain = gale_env.get('GALE_DOMAIN', 'GALE_DOMAIN_UNSET')
		name = name + '@' + gdomain
	
	if callback is None:
		blocker = Blocker()
		cb = lambda k, blocker=blocker: blocker.done_setval(k)
	else:
		cb = callback
	keyobj = sign.Key(name)
	try:
		authcache.find_pubkey(keyobj, lambda k, c=cb: lookup_id_done(k, c))
	except PyGaleErr, e:
		call_error_handler(e)
		keyobj.setpublic(None)
		lookup_id_done(keyobj, cb)
		return
		
	if callback is None:
		while blocker.wait:
			engine.engine.process(authcache.AKD_TIMEOUT)
		return blocker.val

def lookup_id_done(key, callback):
	if key.public() is None:
		callback((None, None))
	else:
		callback((key.name(), key.comment()))

# Look up a list of names.  See lookup_id
def lookup_ids(names, callback=None):
	canonical_names = []
	for name in names:
		if not '@' in name:
			gdomain = gale_env.get('GALE_DOMAIN', 'GALE_DOMAIN_UNSET')
			name = name + '@' + gdomain
			canonical_names.append(name)
	if callback is None:
		blocker = Blocker()
		cb = lambda k, blocker=blocker: blocker.done_setval(k)
	else:
		cb = callback
	keyobjs = map(sign.Key, canonical_names)
	authcache.find_pubkey_list(keyobjs, lambda k, c=cb:
		lookup_ids_done(k, c))
	
	if callback is None:
		while blocker.wait:
			engine.engine.process(authcache.AKD_TIMEOUT)
		return blocker.val

def export_pubkey(keyname, callback = None):
	if callback is None:
		blocker = Blocker()
		cb = lambda k, blocker=blocker: blocker.done_setval(k)
	else:
		cb = callback

	keyobj = sign.Key(keyname)
	# Let's just hope we don't do AKD at this point, OK?
	authcache.find_pubkey_fromdisk(keyobj, lambda k, c=cb:
		export_pubkey2(k, c))
	
	if callback is None:
		while blocker.wait:
			engine.engine.process(0)
		return blocker.val

def export_pubkey2(keyobj, callback):
	callback(keyobj.data())

# ----------------------------------------------------------------------
# Handle non-fatal errors
def set_error_handler(func):
	global ERROR_HANDLER
	ERROR_HANDLER = func

# Notify on progress reports
def set_update_handler(func):
	global UPDATE_HANDLER
	UPDATE_HANDLER = func

def call_error_handler(msg):
	if ERROR_HANDLER is not None:
		ERROR_HANDLER(msg)
def call_update_handler(msg):
	if UPDATE_HANDLER is not None:
		UPDATE_HANDLER(msg)

# ----------------------------------------------------------------------
# Notifications
def notify_in(conn, userid, presence, version=None, instance=None):
	p = Puff()
	p.set_cat(id_category(userid, 'notice', 'login'))
	p.set_text('notice/presence', presence)
	p.set_time('id/time', int(time.time()))
	if version is not None:
		p.set_text('id/class', version)
	if instance is not None:
		p.set_text('id/instance', instance)
	lookup_id(userid, lambda n, id=userid, p=p, c=conn:
		finish_notify_in(n, id, c, p))

def finish_notify_in(namepair, id, conn, p):
	if namepair[0] is None or namepair[1] is None:
		call_update_handler('Unable to lookup %s for presence notification'
			% id)
		return
	p.set_text('message/sender', namepair[1])
	p = p.sign_message(namepair[0])
	if p is not None:
		conn.transmit_puff(p)

def notify_out(conn, userid, presence, version=None, instance=None):
	p = Puff()
	p.set_cat(id_category(userid, 'notice', 'logout'))
	p.set_text('notice/presence', presence)
	if version is not None:
		p.set_text('id/class', version)
	if instance is not None:
		p.set_text('id/instance', instance)
	lookup_id(userid, lambda id, p=p, c=conn: finish_notify_out(id, c, p))


def finish_notify_out(id, conn, p):
	if id[0] == None:
		call_update_handler('Unable to lookup id for notify out message.')
		return
	p.set_text('message/sender', id[1])
	p = p.sign_message(id[0])
	if p is not None:
		conn.transmit_puff(p, will=1)


# ----------------------------------------------------------------------
# Initialization
def init():
#	import os, string
#	if not gale_env.has_key('GALE_SERVER'):
#		print 'Warning: GALE_SERVER not set; using server "localhost"'
#	server = gale_env.get('GALE_SERVER', 'localhost')
#	port = DEFAULT_GALE_PORT
#	if ':' in server:
#		i = string.find(server, ':')
#		if i != -1:
#			port = server[i+1:]
#			server = server[:i]
#	global GALE_SERVER, GALE_PORT
#	GALE_SERVER = server
#	GALE_PORT = port
#	if DEBUG: print 'GALE_SERVER:', GALE_SERVER

	gale_env.init()
	if authcache.DEBUG < DEBUG:
		authcache.DEBUG = DEBUG
	if sign.DEBUG < DEBUG:
		sign.DEBUG = DEBUG

	authcache.init()

def shutdown():
	authcache.shutdown()

# ----------------------------------------------------------------------
# Test Pygale functionality

if __name__ == '__main__':
	opts, args = getopt.getopt(sys.argv[1:], 'rsd:')
	receive = 1
	for (opt, val) in opts:
		if opt == '-r':
			receive = 1
		elif opt == '-s':
			receive = 0
		elif opt == '-d':
			global DEBUG
			DEBUG = int(val)
			print 'Setting DEBUG to', DEBUG
	init()
	if receive:
		testrecv()
	else:
		testsend()
