import time
import pygale, gale_pack, authcache, rsaref

SIG_MAGIC = 'h\023\001\000'
KEY_MAGIC = 'h\023\000\000'
KEY_MAGIC2 = 'h\023\000\002'
KEY_MAGIC3 = 'GALE'
SUBVERSION = '\000\001'
DEBUG = 0

class Key:
	def __init__(self, name=None):
		self.__pubkey = None
		self.__privkey = None
		self.__comment = None
		self.__keyname = name
		self.__data = None
		self.__links_to_me = []
	
	def __repr__(self):
		s = 'Key <%s>' % self.__keyname
		if self.__privkey:
			s = s + ' private'
		if self.__pubkey:
			s = s + ' public'
		if self.__comment:
			s = s + ' [%s]' % self.__comment
		if self.__links_to_me:
			s = s + ' (%s)' % string.join(self.__links_to_me, ',')
		return s
	__str__ = __repr__
	
	# These functions operate on the original name of the key object
	def name(self):
		return self.__keyname
	def setname(self, name):
		self.__keyname = name
	
	def comment(self):
		return self.__comment
	def setcomment(self, comment):
		self.__comment = comment
	
	def public(self):
		return self.__pubkey
	def setpublic(self, pub):
		self.__pubkey = pub

	def private(self):
		return self.__privkey
	def setprivate(self, priv):
		self.__privkey = priv
	
	def trusted(self):
		return self.__trusted
	def settrusted(self, trust):
		self.__trusted = trust
	
	def data(self):
		return self.__data
	def setdata(self, data):
		self.__data = data
	
	def set_links(self, list_of_names):
		self.__links_to_me = list_of_names
	def get_links_to_me(self):
		return self.__links_to_me

# ------------------------------------------------------------
# Routines for signature verification and message decryption
# ------------------------------------------------------------

# Maybe this ought to be called verify() instead
def decode_sig(sig, msg_data, callback):
	if DEBUG > 2: print '--> Starting decode_sig'

	keyobj = Key()
	if DEBUG > 2: print 'Creating keyobj', keyobj

	# _ga_import_sig
	(magic, sig) = gale_pack.pop_data(sig, len(SIG_MAGIC))
	if SIG_MAGIC != magic:
		pygale.call_error_handler("invalid signature format")
		callback(keyobj)
		return
	(siglen, sig) = gale_pack.pop_int(sig)
	if siglen > rsaref.MAX_SIGNATURE_LEN:
		pygale.call_error_handler('invalid signature format')
		callback(keyobj)
		return
	(sigdata, sig) = gale_pack.pop_data(sig, siglen)

	# This is the actual key doing the signing
	(keydata, sig) = gale_pack.pop_data(sig, len(sig))
	keyobj.setdata(keydata)
		
	# Try to find a public key for this signature
	if DEBUG > 2: print 'decode_sig: importing key'
	import_pubkey(keyobj, lambda k, d=sigdata,
		m=msg_data, c=callback: decode_sig2(k, d, m, c))

# Returns the name of the key
# Import a key from a binary representation of the key structure
def import_pubkey(keyobj, callback, trust=0):
	if DEBUG > 2: print '--> Starting import_pubkey'
	keydata = keyobj.data()
	# _ga_import_pub
	# Back up the key to verify its signature, later
	save = keydata
	(magic, keydata) = gale_pack.pop_data(keydata, len(KEY_MAGIC))
	if KEY_MAGIC == magic:
		version = 1
	elif KEY_MAGIC2 == magic:
		version = 2
	elif KEY_MAGIC3 == magic:
		(subversion, keydata) = gale_pack.pop_data(keydata,
			len(SUBVERSION))
		if SUBVERSION == subversion:
			version = 3
		else:
			pygale.call_error_handler('Unsupported version 3 key format')
			keyobj.setpublic(None)
			callback(keyobj)
			return
	else:
		pygale.call_error_handler('Unsupported public key (bad magic)')
		keyobj.setpublic(None)
		callback(keyobj)
		return
	# version must be <= 3 at this point
	if DEBUG >= 3: print 'Public key version:', version

	if version > 1:
		(namelen, keydata) = gale_pack.pop_int(keydata)
		(pubkeyname, keydata) = gale_pack.pop_string(keydata, namelen, chars=1)
		#raise pygale.PyGaleWarn, 'Truncated public key'
	else:	# version == 1
		(pubkeyname, keydata) = gale_pack.pop_nulltermstr(keydata)
	if DEBUG > 2: print 'Found public key named:', pubkeyname
	keyobj.setname(pubkeyname)
	if DEBUG > 2: print 'Setting name in %s to %s' % (`keyobj`, pubkeyname)
	keyobj.settrusted(trust)

	if len(keydata) == 0:
		if DEBUG > 2: print 'Found stub key for', pubkeyname
		stub_done(keyobj, callback)
		return

	# Otherwise, deal with a non-stub key
	# Special-case for different version keys.  Version 3 is the key with
	# fragments in it, which Dan introduced in 0.91b.
	if version == 3:
		fraglist = gale_pack.group_to_FragList(keydata, DAN_IS_STOOPID=1)

		# Verify key if it's signed
		if fraglist.has_key('security/signature'):
			sig = fraglist.get_data('security/signature')
			(sig_len, sig) = gale_pack.pop_int(sig)
			(signature, sig) = gale_pack.pop_data(sig, sig_len)
			msg_data = sig
			new_fragments = gale_pack.group_to_FragList(sig)
			fraglist.update(new_fragments)
		else:
			signature = ''
			# TODO
			# This hacky setting of msg_data is to fix the bug later on
			# where we try to call decode_sig with msg_data.  I need to
			# dig into key formats more to find the right thing to do.
			msg_data = ''

		if fraglist.get_text('key.redirect', None) is not None:
			# Handle a symlink key
			# Gotchas: check for symlink loops (e.g., one links to
			# another, which links to the first)
			symlinkname = fraglist.get_text('key.redirect')
			if DEBUG: print 'Found symlink key to', symlinkname
			links_to_me = keyobj.get_links_to_me()
			if symlinkname in links_to_me:
				# Yikes, there's a loop in the symlink chain!
				pygale.call_error_handler('symlink key loop on key %s' %\
					symlinkname)
				keyobj.setpublic(None)
				callback(keyobj)
				return
			# Otherwise, go fetch it instead
			newkey = Key(symlinkname)
			newkey.set_links(links_to_me + [keyobj.name()])
			authcache.find_pubkey(newkey, callback)
			return

		comment = fraglist.get_text('key.owner', 'Unspecified key.owner')
		time_sign = fraglist.get_time('key.signed', 0)
		time_expire = fraglist.get_time('key.expires', 0)
		rsa_keybits = fraglist.get_int('rsa.bits', 0)
		rsa_modulus = fraglist.get_data('rsa.modulus', '')
		rsa_exponent = fraglist.get_data('rsa.exponent', '')

	# Version 1 or 2 keys are packed binary data.
	elif version == 1 or version == 2:
		if version == 2:
			(comment, keydata) = gale_pack.pop_lenstr(keydata, chars=1)
		else:
			(comment, keydata) = gale_pack.pop_nulltermstr(keydata)
		if DEBUG > 2: print 'Found comment:', comment

		(rsa_keybits, keydata) = gale_pack.pop_int(keydata)
		(rsa_modulus, keydata) = gale_pack.pop_rle(keydata,
			rsaref.MAX_RSA_MODULUS_LEN)
		(rsa_exponent, keydata) = gale_pack.pop_rle(keydata,
			rsaref.MAX_RSA_MODULUS_LEN)

		if version > 1 and keydata:
			(time_sign, keydata) = gale_pack.pop_time(keydata)
			(time_expire, keydata) = gale_pack.pop_time(keydata)
			if DEBUG >= 2: print 'Key signed at:', time_sign.ctime()
			if DEBUG >= 2: print 'Key expires at:', time_expire.ctime()
#			if time_expire <= time.time():
#				pygale.call_error_handler('found expired key for %s' %
#					pubkeyname)
#				callback(None)
#				return
		else:
			# no key timestamp checking in version 1
			pass
		signature = keydata
	else:
		# unsupported version
		pygale.call_error_handler('unsupported key version: %i' %
			version)
		keyobj.setpublic(None)
		callback(keyobj)
		return
	
	# TODO: check for expired keys

	if rsa_keybits < rsaref.MIN_RSA_MODULUS_BITS or \
		rsa_keybits > rsaref.MAX_RSA_MODULUS_BITS:
		pygale.call_error_handler('bad public key bit size %i' % rsa_keybits)
		keyobj.setpublic(None)
		callback(keyobj)
		return

	keyobj.setcomment(comment)
	pubkey = rsaref.R_RSA_PUBLIC_KEY()
	pubkey.bits = rsa_keybits
	pubkey.modulus = rsa_modulus
	pubkey.exponent = rsa_exponent
	keyobj.setpublic(pubkey)

	# Import and validate the signature on this public key
	if not keyobj.trusted():
		if DEBUG > 2: print 'import_pubkey: Verifying sig'
		if version < 3:
			msg_data = save[:-len(keydata)]
		if not signature or not msg_data:
			pygale.call_error_handler('unsigned key %s' % keyobj.name())
			keyobj.setpublic(None)
			callback(keyobj)
			return
		decode_sig(signature, msg_data, lambda k, u=keyobj, c=callback:
			decode_done(k, u, c))

#			decode_sig(sig, frag, lambda key, s=self, p=p, c=callback:
#				s.decode_done(key, p, c))
	else:
		if DEBUG: print 'Key is trusted; not verifying'
		callback(keyobj)
		return

def decode_done(key, upper_key, callback):
	if DEBUG > 2: print 'decode_done: key %s verified key %s' % (key.name(),
		upper_key.name())
	if key.public() is None:
		pygale.call_error_handler('unable to verify signature %s' %
			key.name())
		callback(key)
		return
	callback(upper_key)
	return

def stub_done(key, callback):
	if DEBUG > 2: print 'stub_done: key %s' % key.name()
	# If we get here, then the msg verified ok, and key is a filled-in
	# Key instance
	callback(key)
	return

# The second part of decoding a signature
def decode_sig2(key, sigdata, msg_data, callback):
	# TODO
#	if key.public() is None:
#		pygale.call_error_handler('unable to parse public key')
#		callback(key)
#		return

	if DEBUG > 2: print 'decode_sig2: found key', key.name()
	if key.public() is not None:
		# We have the key right here
		if DEBUG > 2: print 'decode_sig2: have key, verifying'
		if verify_sig(key, sigdata, msg_data):
			if DEBUG > 2: print 'decode_sig2: verification OK'
#			authcache.set_public(key.name(), key.public())
#			authcache.set_comment(key.name(), key.comment())
			callback(key)
			return
	
	# We don't know the public key (or it failed to verify the message),
	# so try the authcache
	cachedkey = authcache.get_key(key.name())
	if cachedkey and cachedkey.public():
		if verify_sig(cachedkey, sigdata, msg_data):
			callback(cachedkey)
			return

	# If it's not in the memory cache, try finding it on disk
	authcache.find_pubkey_fromdisk(key, lambda key, d=sigdata, m=msg_data,
		c=callback: decode_sig3(key, d, m, c))

# The third part of decoding a signature
def decode_sig3(key, sigdata, msg_data, callback):
	if DEBUG > 2: print 'decode_sig3: key', key.name()
	if key.public() is not None:
		# We found it on disk; see if it verifies
		if verify_sig(key, sigdata, msg_data):
			if DEBUG > 2: print 'decode_sig3: verifies OK'
#			authcache.set_public(key.name(), key.public())
#			authcache.set_comment(key.name(), key.comment())
			callback(key)
			return

	# Otherwise, we can't find it (or it doesn't verify the puff)
	# Try (gasp) AKD
	if DEBUG: print 'decode_sig3: uh oh, we have to do AKD'
	authcache.find_pubkey_akd(key, lambda key, d=sigdata, m=msg_data,
		c=callback: decode_sig4(key, d, m, c))

# The fourth step in decoding a signature
def decode_sig4(key, sigdata, msg_data, callback):
	if key.public() is not None:
		if DEBUG: print 'decode_sig4: key found', key.name()
		# We found something via AKD
		if verify_sig(key, sigdata, msg_data):
			# And it verifies!
			if DEBUG: print 'decode_sig4: verifies OK'
#			authcache.set_public(key.name(), key.public())
#			authcache.set_comment(key.name(), key.comment())
			assert key.public() is not None
			authcache.save_pubkey_todisk(key)
			callback(key)
			return
		else:
			key.setpublic(None)
			callback(key)
			return
	else:
		# Otherwise, give up!
		if DEBUG: print 'decode_sig4: no key found'
		pygale.call_error_handler("Warning: no public key for %s" %
			key.name())
		callback(key)
		return

# This returns 1 iff the sigdata is the signed message digest for msg_data,
# signed with the key pubkey
#
# This verifies a single chunk
def verify_sig(key, sigdata, msg_data):
	if DEBUG > 2: print 'verify_sig: verifying sig with key %s (%s)' %\
		(key, key.name())
	if key.trusted():
		if DEBUG > 2: print 'verify_sig: trusted key'
		return 1
# pubkey should never be None at this point
#	if pubkey == None:
#		pygale.call_error_handler("Warning: no public key for %s" % keyname)
#		callback(None)
#		return
	context = rsaref.R_SIGNATURE_CTX()
	retval = context.R_VerifyInit(rsaref.DA_MD5)
	retval = context.R_VerifyUpdate(msg_data, len(msg_data))
	retval = context.R_VerifyFinal(sigdata, len(sigdata), key.public())
	if retval:
		pygale.call_error_handler(
			'signature %s does not verify (key mismatch?)' % key.name())
		return 0
	else:
		return 1
